/*
 *  np_main.c
 *  NtileCodePredictor
 *
 *  Created by Naoto Morikawa on 2/19/07.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-03-05 version 0.1 release
 *  2007-04-10 support the design-protein option, "-d", in main(), 
 *              sort_score_table(), analyse_prediction(), setup_parameters()
 *  2007-04-11 support the check-structure option, "-c", in main(), setup_parameters(),
 *              analyse_prediction()
 *  2007-04-11 support the "-1", "-2", "-3" option in main(), setup_parameters()
 *  2007-04-15 version 0.2 release
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> /* getopt */
#include "np_main.h"
#include "np_pdb_data.h"
#include "np_print.h"

extern int load_protein(const char * filename, struct np_protein * pt_protein);
extern void search_frag_code_table(char * filename,
				   struct np_protein * pt_protein);
extern void print_pred_file(const char * filename,
			    struct np_protein * pt_protein);
extern void create_view_file(const char * filename,
			     struct np_protein * pt_protein);
extern int count_fragment_length(FILE * fp_tbl);

struct np_settings my_settings;


void setup_parameters()
{
	my_settings.search_table_full = FALSE;
	my_settings.search_table5 = FALSE;
	my_settings.search_table7 = FALSE;
	my_settings.search_table9 = FALSE;
	my_settings.search_table15 = FALSE;	// NM 2007-04-11
	my_settings.search_table21 = FALSE;	// NM 2007-04-11
	my_settings.search_table31 = FALSE;	// NM 2007-04-11
	my_settings.compare_enabled = FALSE;
	my_settings.create_view_file = FALSE;
	my_settings.design_protein = FALSE;	// NM 2007-04-10
	my_settings.check_structure = FALSE;	// NM 2007-04-11
	my_settings.tiles_per_res = 0.0;
}

void clear_score_table(struct np_protein * pt_protein)
{
	int chn_ct, aa_ct, tbl_ct;
	
	for (chn_ct = pt_protein->chain_head;
	     chn_ct <= pt_protein->chain_tail; chn_ct++) {
		for (aa_ct = pt_protein->chain[chn_ct].aa_head;
		     aa_ct <= pt_protein->chain[chn_ct].aa_tail;
		     aa_ct++) {
			for (tbl_ct = 0; tbl_ct < SCORE_TABLE_SIZE;
			     tbl_ct++) {
				pt_protein->aa[aa_ct].score_tbl[tbl_ct] = 0.0;
			}
			pt_protein->aa[aa_ct].num_valid_scores = 0.0;
			pt_protein->aa[aa_ct].accuracy = 0.0;
			pt_protein->aa[aa_ct].comp_result = R_NOPREDICTION;
		}
		pt_protein->chain[chn_ct].coverage = 0.0;
	}
	pt_protein->mult_ct = 0;
	pt_protein->success_ct = 0;
	pt_protein->contained_ct = 0;
	pt_protein->hit_ct = 0;
	pt_protein->coverage = 0.0;
	pt_protein->ave_accuracy = 0.0;
	pt_protein->mult_candidate_rate = 0.0;
	pt_protein->success_rate = 0.0;
	pt_protein->extended_success_rate = 0.0;
}

// bubble sort
void sort_score_table(int aa_ind, struct np_protein * pt_protein)
{
	int i, flipped;
	int tmp_ind;
	int table_size;		// NM 2007-04-10
	
	if (my_settings.design_protein == FALSE) {	// NM 2007-04-10
		table_size = ST_UNK;
	} else {
		table_size = L_UNK;
	}
	flipped = 1;
	while (flipped > 0) {
		flipped = 0;
		for (i = 0; i < table_size - 1 ; i++) {
			if (PRED_SCORE(aa_ind, i) < PRED_SCORE(aa_ind, i + 1)) {
				if (0) fprintf(STDOUT, "sort_score_table() : DEBUG %d (%f) < %d (%f)\n",
						   PRED_CODE(aa_ind, i),
						   PRED_SCORE(aa_ind, i),
						   PRED_CODE(aa_ind, i + 1),
						   PRED_SCORE(aa_ind, i + 1));
				flipped = 1;
				tmp_ind = PRED_CODE(aa_ind, i);
				PRED_CODE(aa_ind, i) = PRED_CODE(aa_ind, i + 1);
				PRED_CODE(aa_ind, i + 1) = tmp_ind;
			}
		}
	}
}

void initialize_predictor_table(int aa_ind, struct np_protein * pt_protein)
{
	int tbl_ct;
	
	for (tbl_ct = 0; tbl_ct < SCORE_TABLE_SIZE; tbl_ct++) {
		pt_protein->aa[aa_ind].prediction[tbl_ct] = tbl_ct;
	}
}

// Sort ntile code in descending order based on the score
void sort_ntile_code(struct np_protein * pt_protein)
{
	int chn_ct, aa_ct;
	
	for (chn_ct = pt_protein->chain_head;
	     chn_ct <= pt_protein->chain_tail; chn_ct++) {
		for (aa_ct = pt_protein->chain[chn_ct].aa_head;
		     aa_ct <= pt_protein->chain[chn_ct].aa_tail;
		     aa_ct++) {
			initialize_predictor_table(aa_ct, pt_protein);
			if (0) print_prediction(aa_ct, pt_protein);
			sort_score_table(aa_ct, pt_protein);
			if (0) print_prediction(aa_ct, pt_protein);
		}
	}
}

void analyse_prediction(struct np_protein * pt_protein)
{
	int chn_ct, aa_ct;
	int tbl_ct;
	int chn_hit, total_hit;
	int chn_aa_num, total_aa_num;
	int num_topscorer;
	float max_score;
	int i;
	int cutoff_num;
	
	if (my_settings.check_structure == TRUE) {	// NM 2007-04-11
		cutoff_num = ((int) my_settings.tiles_per_res - 1) / 2;
	} else if (my_settings.design_protein == TRUE) {
		cutoff_num = 0;
	} else {		// NM 2007-04-10
		cutoff_num = 2;	// Discard two amino acids at both ends of the chain
	}
	for (chn_ct = pt_protein->chain_head;
	     chn_ct <= pt_protein->chain_tail; chn_ct++) {
		aa_ct = pt_protein->chain[chn_ct].aa_head;
		for (i = 0; i < cutoff_num; i++) {
			pt_protein->aa[aa_ct + i].num_valid_scores = 0;
		}
		aa_ct = pt_protein->chain[chn_ct].aa_tail;
		for (i = 0; i < cutoff_num; i++) {
			pt_protein->aa[aa_ct - i].num_valid_scores = 0;
		}
	}
		
	total_hit = 0;
	total_aa_num = 0;
	for (chn_ct = pt_protein->chain_head;
	     chn_ct <= pt_protein->chain_tail; chn_ct++) {
		chn_hit = 0;
		for (aa_ct = pt_protein->chain[chn_ct].aa_head;
		     aa_ct <= pt_protein->chain[chn_ct].aa_tail;
		     aa_ct++) {
			num_topscorer = 0;
			if (pt_protein->aa[aa_ct].num_valid_scores > 0) {
				chn_hit++;
				num_topscorer++;
				max_score = PRED_SCORE(aa_ct, 0);
				for (tbl_ct = 1; tbl_ct < ST_UNK - 1; tbl_ct++) {
					if (FLOAT_NEQ(PRED_SCORE(aa_ct, tbl_ct) - max_score))
						break;
					num_topscorer++;
				}
				pt_protein->aa[aa_ct].accuracy =
					max_score / pt_protein->aa[aa_ct].num_valid_scores;
			}
			pt_protein->aa[aa_ct].num_topscorer = num_topscorer;
		}
		chn_aa_num = pt_protein->chain[chn_ct].aa_num - cutoff_num * 2;	// NM 2007-04-11
		if (chn_aa_num <= 0) {	// NM 2007-04-10
			fprintf(STDERR, "analyse_prediction() : CAUTION pt_protein->chain[%d].aa_num = %d (chn_aa_num = %d)\n",
				chn_ct, pt_protein->chain[chn_ct].aa_num, chn_aa_num);
			chn_aa_num = 1;
		}
		pt_protein->chain[chn_ct].coverage =
			(float) chn_hit / (float) chn_aa_num;
		total_hit += chn_hit;
		total_aa_num += chn_aa_num;
	}
	pt_protein->hit_ct = total_hit;
	pt_protein->coverage = (float) total_hit / (float) total_aa_num;
}

/**
 * main - encodes the structure of a protein into a binary sequence.
 * @argc   :
 * @argv[argc-1] : input amino acid sequence file
 *
 * SYNOPSIS
 *        NtileCodePredictor filename
 *
 * INPUT
 *	Amino acid sequences of a protein (".ent", ".code.", or ".seq" file)
 *	Fragment to Ntile code conversion table (".tbl" files)
 *
 * OUTPUT
 *	Predicated Ntile code of the sequence (".pred" file)
 */

#define FRAG_CODE_TABLE_FULL "frag_code5_full.tbl"
#define FRAG_CODE_TABLE5 "frag_code5.tbl"
#define FRAG_CODE_TABLE7 "frag_code7.tbl"
#define FRAG_CODE_TABLE9 "frag_code9.tbl"
#define FRAG_CODE_TABLE5_FULL "frag_code5_full.tbl"	// NM 2007-04-10
#define FRAG_CODE_TABLE7_FULL "frag_code7_full.tbl"	// NM 2007-04-10
#define FRAG_CODE_TABLE9_FULL "frag_code9_full.tbl"	// NM 2007-04-10
#define FRAG_CODE_TABLE15_FULL "frag_code15_full.tbl"	// NM 2007-04-11
#define FRAG_CODE_TABLE21_FULL "frag_code21_full.tbl"	// NM 2007-04-11
#define FRAG_CODE_TABLE31_FULL "frag_code31_full.tbl"	// NM 2007-04-11

int main (int argc, char ** argv)
{
	const char * protein_file;
	struct np_protein my_np_protein;
	int c;
	char * frag_code_table = NULL;
	BOOLEAN tbl_specified = FALSE;
	
	// Setup the score table of Ntile code for each amino acid.
	setup_parameters();
	while ((c = getopt(argc, argv, "f579123vhdct:")) != -1 ) {	// NM 2007-04-11
		switch (c) {
			case 'f':
				my_settings.search_table_full = TRUE;
				tbl_specified = TRUE;
				my_settings.tiles_per_res = 5.0;
				break;
			case '5':
				my_settings.search_table5 = TRUE;
				tbl_specified = TRUE;
				my_settings.tiles_per_res = 5.0;
				break;
			case '7':
				my_settings.search_table7 = TRUE;
				tbl_specified = TRUE;
				if (my_settings.tiles_per_res == 0.0)
					my_settings.tiles_per_res = 7.0;
				break;
			case '9':
				my_settings.search_table9 = TRUE;
				tbl_specified = TRUE;
				if (my_settings.tiles_per_res == 0.0)
					my_settings.tiles_per_res = 9.0;
				break;
			case '1':	// NM 2007-04-11
				my_settings.search_table15 = TRUE;
				tbl_specified = TRUE;
				my_settings.tiles_per_res = 15.0;
				break;
			case '2':	// NM 2007-04-11
				my_settings.search_table21 = TRUE;
				tbl_specified = TRUE;
				my_settings.tiles_per_res = 21.0;
				break;
			case '3':	// NM 2007-04-11
				my_settings.search_table31 = TRUE;
				tbl_specified = TRUE;
				my_settings.tiles_per_res = 31.0;
				break;
			case 'v':
				my_settings.create_view_file = TRUE;
				break;
			case 'd':	// NM 2007-04-10
				my_settings.design_protein = TRUE;
				break;
			case 'c':	// NM 2007-04-11
				my_settings.check_structure = TRUE;
				my_settings.design_protein = TRUE;
				break;
			case 't':
				frag_code_table = optarg;
				if (my_settings.tiles_per_res == 0.0) {
					FILE * fp_frag_tbl;
					
					fp_frag_tbl = fopen(frag_code_table, "r");
					if (fp_frag_tbl == NULL) {
						fprintf(STDERR, "main() : Can't open %s\n", frag_code_table);
						return 1;
					}
					my_settings.tiles_per_res =
						(float) count_fragment_length(fp_frag_tbl);
					if (fclose(fp_frag_tbl)) {
						fprintf(STDERR, "main() : Can't close %s\n", frag_code_table);
						abort();
					}
				}
				tbl_specified = TRUE;
				break;
			case 'h':	// help
				fprintf(STDERR, "Synopsis : NtileCodePredictor [-f][-5][-7][-9][-1][-2][-3][-v][-d][-c][-t table_name][-h] protein_file_name\n");
				return 0;
			default:
				fprintf(STDERR, "main() : Unrecognised option : -%c\n", optopt);
				return 1;
		}
	}
	if (tbl_specified == FALSE) {
		my_settings.search_table_full = TRUE;	// default
		my_settings.tiles_per_res = 5.0;
	}
	
	// Load the target protein
	protein_file = argv[argc -1];
	fprintf(STDOUT, "INPUT SEQUENCE FILE : %s\n", protein_file);
	my_np_protein.filename = protein_file;
	if (load_protein(protein_file, &my_np_protein) == FAIL) {
		fprintf(STDERR, "main() : ERROR Load protein file failed\n");
		return 1;
	}
	
	// For each amino acid fragment of length N, 
	//	search the "frag_code.tbl" for the fragment.
	//	If hit, renew the score table.
	clear_score_table(&my_np_protein);
	if (0) print_np_protein(&my_np_protein, FALSE, FALSE);
	
	if (frag_code_table != NULL) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", frag_code_table);
		search_frag_code_table(frag_code_table, &my_np_protein);
	}
	if (my_settings.search_table_full == TRUE) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE_FULL);
		search_frag_code_table(FRAG_CODE_TABLE_FULL, &my_np_protein);
	}
	if (my_settings.design_protein == FALSE) {	// NM 2007-04-10
		if (my_settings.search_table5 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE5);
			search_frag_code_table(FRAG_CODE_TABLE5, &my_np_protein);
		}
		if (my_settings.search_table7 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE7);
			search_frag_code_table(FRAG_CODE_TABLE7, &my_np_protein);
		}
		if (my_settings.search_table9 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE9);
			search_frag_code_table(FRAG_CODE_TABLE9, &my_np_protein);
		}
	} else {
		if (my_settings.search_table5 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE5_FULL);
			search_frag_code_table(FRAG_CODE_TABLE5_FULL, &my_np_protein);
		}
		if (my_settings.search_table7 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE7_FULL);
			search_frag_code_table(FRAG_CODE_TABLE7_FULL, &my_np_protein);
		}
		if (my_settings.search_table9 == TRUE) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE9_FULL);
			search_frag_code_table(FRAG_CODE_TABLE9_FULL, &my_np_protein);
		}
	}
	if (my_settings.search_table15 == TRUE) {		// NM 2007-04-11
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE15_FULL);
		search_frag_code_table(FRAG_CODE_TABLE15_FULL, &my_np_protein);
	}
	if (my_settings.search_table21 == TRUE) {		// NM 2007-04-11
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE21_FULL);
		search_frag_code_table(FRAG_CODE_TABLE21_FULL, &my_np_protein);
	}
	if (my_settings.search_table31 == TRUE) {		// NM 2007-04-11
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "FRAG_CODE_TABLE : %s\n", FRAG_CODE_TABLE31_FULL);
		search_frag_code_table(FRAG_CODE_TABLE31_FULL, &my_np_protein);
	}
	
	if (0) print_np_protein(&my_np_protein, TRUE, FALSE);

	sort_ntile_code(&my_np_protein);
	analyse_prediction(&my_np_protein);
	if (DEBUG) print_np_protein(&my_np_protein, FALSE, TRUE);
	
	// Output
	print_pred_file(protein_file, &my_np_protein);
	
	// View file
	if (my_settings.create_view_file == TRUE)
		create_view_file(protein_file, &my_np_protein);
	
	//free memory
	if (my_np_protein.aa != NULL) free(my_np_protein.aa);
	if (my_np_protein.chain != NULL) free(my_np_protein.chain);
	

	return 0;
}
