/*
 *  np_main.h
 *  NtileCodePredictor
 *
 *  Created by Naoto Morikawa on 2/19/06.
 *  Copyright (C) 2007 Naoto Morikawa. All rights reserved.
 *
 *  2007-03-05 version 0.1 release
 *  2007-04-15 version 0.2 release
 *
 */

/*  [MEMO] Data structures of the PDB records
 *  Molecules are organized as follows:
 *    ----------------------------------
 *    Chain A    Molecule 1    Model 1
 *    --------
 *    Chain B
 *    ---------------------
 *    Chain C    Molecule 2
 *    --------
 *    Chain D
 *    ----------------------------------
 *    Chain A    Molecule 1    Model 2
 *    --------
 *    Chain B
 *    ---------------------
 *    Chain C    Molecule 2
 *    --------
 *    Chain D
 *    ----------------------------------
 */

#define LINE_WIDTH 80		// the line width of ".ent" files

/**
 * struct pe_pdb_line - is a line of PDB file
 * @record_name[6] : record name
 * @record_body[74]: record body
 *
 * Each line in the PDB file is self-identifying.
 * And each record consists of one or more lines.
 */
struct pe_pdb_line {
	char record_name[6];	/* column 1-6 */
	char record_body[74];	/* column 6-80 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_header - is the HEADER record
 * @record_name[6]: record name
 * @class[40]     : class of the molecule(s)
 * @date[9]       : deposition date
 * @idcode[4]     : PDB identifier
 */
struct pe_pdb_header {
	char record_name[6];	/* column 1-6 */
	char blank1[4];		/* column 7-10 */
	char class[40];		/* column 11-50 */
	char date[9];		/* column 51-59 */
	char idcode[4];		/* column 63-66 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_compnd - is the COMPND record
 * @record_name[6] : record name
 * @continuation[2]: continuation (for concatenation of multiple records)
 * @spec_list[60]  : deposition date
 */
struct pe_pdb_compnd {
	char record_name[6];	/* column 1-6 */
	char brank1[2];
	char continuation[2];	/* column 9-10 */
	char spec_list[60];	/* column 11-70 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_model - is the MODEL record
 * @record_name[6]: record name
 * @serial_num[4] : model serial number
 */
struct pe_pdb_model {
	char record_name[6];	/* column 1-6 */
	char blank[4];
	char serial_num[4];	/* column 11-14 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_atom - is the ATOM record
 * @record_name[6]: record name
 * @serial_num[4] : atom serial number
 * @atom_name[4]  : atom name
 * @alt_loc       : alternate location indicator
 * @res_name[3]   : residue name
 * @chain_id      : chain identifier
 * @res_seq[4]    : residue sequence identifier
 * @i_code        : code for insertion of residues
 * @x[8]          : orthogonal coordinates for x in angstroms, real(8.3)
 * @y[8]          : orthogonal coordinates for y in angstroms, real(8.3)
 * @z[8]          : orthogonal coordinates for z in angstroms, real(8.3)
 * @occupancy[6]  : occupancy, real(6.2)
 * @temp_fac[6]   : temperature factor, real(6.2)
 * @seg_id[4]     : segment identifier
 * @element[2]    : element symbol
 * @charge[2]     : charge on the atom
 */
struct pe_pdb_atom {
	char record_name[6];	/* column 1-6 */
	char serial_num[5];	/* column 7-11 */
	char blank1;		/* column 12 */
	char atom_name[4];	/* column 13-16 */
	char alt_loc;		/* column 17 */
	char res_name[3];	/* column 18-20 */
	char blank2;		/* column 21 */
	char chain_id;		/* column 22 */
	char res_seq[4];	/* column 23-26 */
	char i_code;		/* column 27 */
	char blank3[3];		/* column 28-30 */
	char x[8];		/* coloum 31-38  Real(8.3) */
	char y[8];		/* coloum 39-46 */
	char z[8];		/* coloum 47-54 */
	char occupancy[6];	/* coloum 55-60  Real(6.2) */
	char temp_fac[6];	/* coloum 61-66 */
	char blank4[6];		/* column 67-72 */
	char seg_id[4];		/* coloum 73-76 */
	char element[2];	/* coloum 77-78 */
	char charge[2];		/* coloum 79-80 */
	char pad;		/* room for the newline character */
};

/* [TODO] Support the HETATM record! */

/**
 * struct pe_pdb_ter - is the TER record
 * @record_name[6]   : record name
 * @serial_num[4]    : serial number
 * @res_name[3]      : residue name
 * @chain_id         : chain identifier
 * @res_serial_num[4]: residue sequence number
 * @i_code           : insetion code
 */
struct pe_pdb_ter {
	char record_name[6];	/* column 1-6 */
	char serial_num[5];	/* column 7-11 */
	char blank1[6];		/* column 12-17 */
	char res_name[3];	/* column 18-20 */
	char blank2;		/* column 21 */
	char chain_id;		/* column 22 */
	char res_serial_num[4];	/* column 23-26 */
	char i_code;		/* column 27 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_endmdl - is the ENDMDL record
 * @record_name[6]: record name
 */
struct pe_pdb_endmdl {
	char record_name[6];	/* column 1-6 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_master - is the MASTER record
 * @record_name[6]: record name
 * @num_remark[5] : number of REMARK records
 * @num_het[5]    : number of HET records
 * @num_helix[5]  : number of HELIX records
 * @num_sheet[5]  : number of SHEET records
 * @num_turn[5]   : number of TURN records
 * @num_site[5]   : number of SITE records
 * @num_xform[5]  : number of ORIGX, SCALE, and MTRIX records
 * @num_coor[5]   : number of ATOM and HETATM records
 * @num_ter[5]    : number of TER records
 * @num_conect[5] : number of CONECT records
 * @num_seqres[5] : number of SEQRES records
 */
struct pe_pdb_master {
	char record_name[6];	/* column 1-6 */
	char blank1[4];		/* column 7-10 */
	char num_remark[5];	/* column 11-15 */
	char zeros[5];		/* column 16-20 */
	char num_het[5];	/* column 21-25 */
	char num_helix[5];	/* column 26-30 */
	char num_sheet[5];	/* column 31-35 */
	char num_turn[5];	/* column 36-40 */
	char num_site[5];	/* column 41-45 */
	char num_xform[5];	/* column 46-50 */
	char num_coor[5];	/* column 51-55 */
	char num_ter[5];	/* column 56-60 */
	char num_conect[5];	/* column 61-65 */
	char num_seqres[5];	/* column 66-70 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_pdb_end - is the END record
 * @record_name[6]: record name
 */
struct pe_pdb_end {
	char record_name[6];	/* column 1-6 */
	char pad;		/* room for the newline character */
};

/**
 * struct pe_counter - is a counter
 *
 */
struct pe_counter {
	int model;
	int molecule;
	int chain;
	int residue;
	int sequence;
	int atom;
};
